/* Copyright 2022 Pierre Ossman for Cendio AB
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef __FL_NAVIGATION_H__
#define __FL_NAVIGATION_H__

#include <FL/Fl_Group.H>

class Fl_Scroll;

class Fl_Navigation: private Fl_Group {
public:
  Fl_Navigation(int x, int y, int w, int h);
  ~Fl_Navigation();

  Fl_Widget *value();
  int value(Fl_Widget*);

  void client_area(int &rx, int &ry, int &rw, int &rh, int lw);

  void draw() override;

  // Delegation to underlying widget
  void begin();
  void end();

private:
  void update_labels();

  static void label_pressed(Fl_Widget *widget, void *user_data);

private:
  Fl_Scroll *scroll;
  Fl_Group *labels;
  Fl_Group *pages;
};

#endif
